/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */
#include <ncurses.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <netdb.h>
#include <assert.h>
#include <sys/stat.h>
#include <limits.h>

#include "main.h"

#include "curs_color.h"
#include "curs_lib.h"
#include "keymap.h"
#include "lib.h"
#include "misc.h"
#include "options.h"

void cmd_redraw_screen(int *menu_redraw)
{
	TURN_ON_FLAG(*menu_redraw, REDRAW_FULL);
}

/*
 * flags must be bitwise OR'ed with one or more of the following
 * (you'll definitely want at least one of those flags):
 * HALF_FILLED - path is half-filled with data already
 * ABSOLUTE - path must be absolute and start with ~ or /
 * MUST_EXIST - path must exist
 * DIRECTORY - path must be a dir or a symbolic link to one, can only
 * be specified if MUST_EXIST is given
 * REG_FILE - if MUST_EXIST is given, path must be a regular file or a
 * symbolic link to one
 * REGISTERED_EXT - the same as REG_FILE, also, path must end with one of
 * the extensions defined in the OPT_FILENAME_EXTENSIONS option
 * ALLOW_BROWSE - allow entering the file system browser
 */
char *enter_path(char *path, size_t pathlen, int flags, const char *fmt,...)
{
    struct stat statbuf;
    char prompt[STRING];
    va_list ap;
    
    va_start(ap, fmt);
    vsnprintf(prompt, sizeof(prompt), fmt, ap);
    va_end(ap);

    if (flags & HALF_FILLED) {
	if (enter_string(path, pathlen, HALF_FILLED, ALLOW_BROWSE, prompt) == NULL)
	    return NULL;
	else if (Browse == TRUE)
	    return path; /* the callee will check `Browse' on return and ignore path */
    } else {
	if (enter_string(path, pathlen, EMPTY, ALLOW_BROWSE, prompt) == NULL || pathlen < 2)
	    return NULL;
	else if (Browse == TRUE)
	    return path;	/* the callee will check `Browse' on return and ignore path */
    }

    if (flags & ABSOLUTE) {
	if (path[0] != '/' && path[0] != '~') {
	    msg("Path must start with tilde (~) or slash (/)).");
	    return NULL;
	}
    }
    if (path[0] == '.') {
	if (path[1] == '/') {
	    memmove(path, path + 2, strlen(path) + 2 + 1);
	    pathlen -= 2;
	} else if (path[1] == '.') {
	    msg("You can't specify paths starting with dot-dot (..)");
	    return NULL;
	}
    }
    if (path[0] == '~') {
	if (expand_path(path, pathlen) == NULL) {
	    msg("Can't expand %s.", path);
	    return NULL;
	}
    }
    if (flags & MUST_EXIST) {
	if (access(path, F_OK) != 0) {
	    perror_msg("Can't access %s", path);
	    return NULL;
	}
	if ((flags & DIRECTORY) || (flags & REG_FILE) || (flags & REGISTERED_EXT)) {
	    if (stat(path, &statbuf) < 0) {
		perror_msg("Can't stat %s", path);
		return NULL;
	    }
	    if (flags & DIRECTORY) {
		if (!S_ISDIR(statbuf.st_mode)) {
		    msg("%s is not a directory.", path);
		    return NULL;
		}
	    } else if ((flags & REG_FILE) || (flags & REGISTERED_EXT)) {
		if (!S_ISREG(statbuf.st_mode)) {
		    msg("%s is not a regular file.", path);
		    return NULL;
		}
	    } else
		assert(0);
	}
    }
    if (flags & REGISTERED_EXT) {
	if (has_registered_filename_extension(path) == FALSE) {
	    MSG_FILENAME_EXTENSION_NOT_REGISTERED(path);
	    return NULL;
	}
    } return path;
}

/*
 * The same function as perror(3), except that this function takes a variable
 * number of arguments and prints the output on the message line.
 */
void perror_msg(const char *fmt,...)
{
    char msg[STRING];
    int saved_errno = errno;
    char *perror_str = strerror(saved_errno);
    va_list ap;

    va_start(ap, fmt);
    vsnprintf(msg, sizeof(msg), fmt, ap);
    va_end(ap);

    snprintf(msg + strlen(msg), sizeof(msg) - strlen(msg),
	     ": %s (errno = %d).",
	     perror_str ? perror_str : "unknown error", saved_errno);

    attrset(msg_line_attr);
    mvaddnstr(MSG_LINE, 0, msg, COLS);
    clrtoeol();
}

/*
 * Print output on the message line as specified by `fmt'.
 */
void msg(const char *fmt,...)
{
    char msg[STRING];
    va_list ap;

    va_start(ap, fmt);
    vsnprintf(msg, sizeof(msg), fmt, ap);
    va_end(ap);

    attrset(msg_line_attr);
    mvaddnstr(MSG_LINE, 0, msg, COLS);
    clrtoeol();
}

/*
 * Print output as specified by `fmt', followed by [y] if `def' is true or [n]
 * if `def' is false, on the bottom line. Function returns when the user hits
 * y, n or enter. If enter is hit, the default value is assumed.
 * Return TRUE if answer is yes, FALSE if answer is no.
 */
int yes_or_no(bool def, const char *fmt,...)
{
    int ch;
    char msg[STRING];
    char buf[STRING];
    va_list ap;
    
    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
    va_end(ap);

    strfcpy(msg, buf,
	    COLS < sizeof(buf)
	    ? COLS - strlen(" [c] ")
	    : sizeof(buf) - strlen(" [c] ")
	    );
    sprintf(msg + strlen(msg), " [%c] ", def ? 'y' : 'n');

    attrset(msg_line_attr);
    mvaddstr(MSG_LINE, 0, msg);
    clrtoeol();
    refresh();

    while (1) {
	ch = getch();
	if (ch == '\n') {
	    break;
	} else if (ch == 'y') {
	    def = TRUE;
	    break;
	} else if (ch == 'n') {
	    def = FALSE;
	    break;
	}
    }

    addstr(def ? "Yes" : "No");
    refresh();
    return def;
}

/*
 * This is very much the same as menu_get_command().
 * Get and parse command from the user.
 * Do not return before a valid command or a printable character is input.
 * If a valid command is received, set `cmd' to the identification number
 * of it and return 0, else if a printable character is received set `cmd'
 * to 0 and return the character.
 */
static int enter_string_get_command(int *cmd)
{
    int key;
    struct keymap_entry *p;

    while ((key = getch()) == ERR)
	;
    (void) ungetch(key);

 start_over_again:

    key = getch();
    /* Two `ESC's in a row mean cancel the current command */
    if (key == ESC) {
	key = getch();
	if (key == ESC)
	    goto start_over_again;
	(void) ungetch(key);
	key = ESC;
    }

    /* Scan through the keymap, looking for matching entries */
    p = keymap;
    while (p->cmd_id != KEYMAP_TERMINATOR) {
	/* If the keymap entry is valid for the line editor */
	if (p->menus & LINE_EDITOR) {
	    if (p->firstchar == key) {
		if (p->nextchar == 0) {
		    *cmd = p->cmd_id;
		    return 0;
		} else {
		    break;
		}
	    }
	}
	p++;
    }
    /* If `key' is not equal to the first character of any keymap entries */
    if (p->cmd_id == KEYMAP_TERMINATOR) {
	if (isprint(key)) {
	    *cmd = 0;
	    return key;
	} else
	    goto start_over_again;
    }

    /* Two `ESC's in a row mean cancel the current command */
    key = getch();
    if (key == ESC) {
	key = getch();
	if (key == ESC)
	    goto start_over_again;
	(void) ungetch(key);
	key = ESC;
    }

    /* Scan through the keymap, looking for matching entries */
    p = keymap;
    while (p->cmd_id != KEYMAP_TERMINATOR) {
	if ((p->menus & LINE_EDITOR) && p->nextchar == key) {
	    *cmd = p->cmd_id;
	    return 0;
	}
	p++;
    }

    /* Unknown command */
    goto start_over_again;
}

#define REDRAW_INIT 1		/* recalculate lengths */
#define REDRAW_LINE 2		/* redraw entire line */
#define REDRAW_EOL 3		/* redraw from current position to eol */
#define REDRAW_PREV_EOL 4	/* redraw from curpos-1 to eol */

/*
 * 1. Print output on the message line as specified by `fmt'.
 * 2. Get input from the user into `buf' (not including the linefeed ending
 *    the input) and \0 terminate it.
 * 3. Clear the message line and refresh the screen.
 * Return a pointer to `buf', or NULL if the user aborts (`CMD_ABORT'). If
 * NULL is returned, the content of `buf' is unknown (and probably not \0
 * terminated). Be aware that an empty string can be returned (which is the
 * case when you only give the function a linefeed).
 */
char *enter_string(char *buf, size_t buflen, int bufstate, bool allow_browse,
		   const char *fmt,...)
{
    int curpos = 0;		/* the location of the cursor */
    int lastchar = 0;		/* offset of the last char in the string */
    int begin = 0;		/* first character displayed on the line */
    int ch = 0;			/* last typed character */
    int width;			/* width of field */
    int redraw = REDRAW_INIT;	/* when/what to redraw */
    int cmd;                    /* command given by the user */
    int j;
    int x = -1;
    int y = -1;
    char prompt[STRING];
    int prompt_len;
    va_list ap;

    /* global variable which will be set to true when CMD_BROWSE is hit */
    Browse = FALSE;

    if (bufstate == EMPTY)
	buf[0] = '\0';
    /* else the value of bufstate is `HALF_FILLED', which means it's
       nul terminated */

    va_start(ap, fmt);
    vsnprintf(prompt, sizeof(prompt), fmt, ap);
    va_end(ap);

    attrset(msg_line_attr);
    prompt_len = strlen(prompt) > COLS - 10 ? COLS - 10 : strlen(prompt);
    mvaddnstr(MSG_LINE, 0, prompt, prompt_len);
    clrtoeol();
    y = MSG_LINE;
    x = prompt_len;
    width = COLS - x - 1;

    while (1) {

	if (redraw) {
	    if (redraw == REDRAW_INIT) {
		/* full redraw */
		lastchar = curpos = strlen((char *) buf);
		begin = lastchar - width;
	    }
	    if (begin < 0)
		begin = 0;

	    switch (redraw) {
	    case REDRAW_PREV_EOL:
		j = curpos - 1;
		break;
	    case REDRAW_EOL:
		j = curpos;
		break;
	    default:
		j = begin;
	    }

	    move(y, x + j - begin);
	    for (; j < lastchar && j < begin + width; j++)
		addch(buf[j]);
	    clrtoeol();
	    if (redraw != REDRAW_INIT)
		move(y, x + curpos - begin);

	    redraw = 0;
	}
	refresh();

	ch = enter_string_get_command(&cmd);

	switch (cmd) {

	case CMD_ABORT:
	    CLEARLINE(MSG_LINE);
	    refresh();
	    return NULL;

	case CMD_BROWSE:
	    if (allow_browse) {
		Browse = TRUE;
		strfcpy(buf, "Don't use me!", buflen);
		return buf;
	    } else
		break;

	case CMD_BACKWARD_DELETE_CHAR:
	    if (curpos == 0)
		break;
	    for (j = curpos; j < lastchar; j++)
		buf[j - 1] = buf[j];
	    curpos--;
	    lastchar--;
	    if (curpos > begin) {
		if (lastchar == curpos) {
		    move(y, x + curpos - begin);
		    delch();
		} else
		    redraw = REDRAW_EOL;
	    } else {
		begin -= width / 2;
		redraw = REDRAW_LINE;
	    }
	    break;

	case CMD_BEGINNING_OF_LINE:
	    /* reposition the cursor at the begininning of the line */
	    curpos = 0;
	    if (begin) {
		/* the first char is not displayed, so readjust */
		begin = 0;
		redraw = REDRAW_LINE;
	    } else
		move(y, x);
	    break;

	case CMD_END_OF_LINE:
	    curpos = lastchar;
	    if (lastchar < begin + width)
		move(y, x + lastchar - begin);
	    else {
		begin = lastchar - width / 2;
		redraw = REDRAW_LINE;
	    }
	    break;

	case CMD_KILL_WHOLE_LINE:
	    curpos = lastchar = 0;
	    redraw = REDRAW_LINE;
	    break;

	case CMD_KILL_LINE:
	    lastchar = curpos;
	    clrtoeol();
	    break;

	case CMD_BACKWARD_CHAR:
	    if (curpos != 0) {
		curpos--;
		if (curpos < begin) {
		    begin -= width / 2;
		    redraw = REDRAW_LINE;
		} else
		    move(y, x + curpos - begin);
	    }
	    break;

	case CMD_FORWARD_CHAR:
	    if (curpos != lastchar) {
		curpos++;
		if (curpos >= begin + width) {
		    begin = curpos - width / 2;
		    redraw = REDRAW_LINE;
		} else
		    move(y, x + curpos - begin);
	    }
	    break;

	case CMD_DELETE_CHAR:
	    if (curpos != lastchar) {
		for (j = curpos; j < lastchar; j++)
		    buf[j] = buf[j + 1];
		lastchar--;
		redraw = REDRAW_EOL;
	    }
	    break;

	case CMD_BACKWARD_KILL_WORD:
	    /* words are composed of letters and digits,
	     we do what Emacs does */
	    if (curpos != 0) {
		j = curpos;
		while (j > 0 && !isalnum(buf[j - 1]))
		    j--;
		while (j > 0 && isalnum(buf[j - 1]))
		    j--;
		ch = j;	/* save current position */

		while (curpos < lastchar)
		    buf[j++] = buf[curpos++];
		lastchar = j;
		curpos = ch;	/* restore current position */
		/* update screen */
		if (curpos < begin) {
		    begin = curpos - width / 2;
		    redraw = REDRAW_LINE;
		} else
		    redraw = REDRAW_EOL;
	    }
	    break;

	case CMD_ENTER:
	    CLEARLINE(MSG_LINE);
	    refresh();
	    buf[lastchar] = '\0';
	    /* note that empty strings can be input and returned
	       (when `lastchar' is 0) */
	    return buf;

	default:
	    /* The key hit is neither a command or enter */

	    /* Check if there's a missing case for `cmd' */
	    assert(cmd == 0);
	    /* enter_string_get_command() must only return printable characters */
	    assert(isprint(ch));
	    for (j = lastchar; j > curpos; j--)
		buf[j] = buf[j - 1];
	    buf[curpos++] = ch;
	    lastchar++;

	    if (curpos >= begin + width) {
		begin = curpos - width / 2;
		redraw = REDRAW_LINE;
	    } else if (curpos == lastchar)
		addch(ch);
	    else
		redraw = REDRAW_PREV_EOL;
	}
    }

    /* not reached */
}
